/* Copyright (C) 2001 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@msu.edu> on the Internet,
   or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA through
   more mundane means.  
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "test.h"
#include "uniformity.h"

/* Command-line options. */
struct options
  {
    int dims;
    int count;
    int average;
    double sig;

    enum dist
      {
	DIST_UNIFORM,
	DIST_DECREASING,
	DIST_NORMAL
      }
    dist;

    enum mst
      {
	MST_PRIM_BIN_HEAP,
	MST_PRIM_FIB_HEAP
      }
    mst;

    enum rng
      {
	RNG_MT,
	RNG_STD
      }
    rng;

    double prox;
    int runs;
    unsigned seed;
  };

const char *pgm_name = "unf-test";

/* Print helpful syntax message and exit. */
static void 
usage (void) 
{
  printf ("unf-test, test program for uniformity modules.\n\n"
	  "Usage: %s [OPTION]...\n\n"
	  "In the option descriptions below, CAPITALS denote arguments.\n"
	  "If a long option shows an argument as mandatory, then it is\n"
	  "mandatory for the equivalent short option also.\n\n"
	  "-d, --dims=DIMS       Number of dimensions in test (default 5).\n"
	  "-c, --count=N         Number of points to test (default 100).\n"
	  "-a, --average=AVG     Number of trials to average (default 1).\n"
	  "-S, --sig=SIG         Significance level (default .05).\n"
	  "-D, --dist=DIST       Distribution to test.  DIST is one of:\n"
	  "                        uniform     uniform dist (default)\n"
	  "                        decreasing  X1 > X2 > ... > Xn-1 > Xn\n"
	  "                        normal      normal distribution\n"
	  "-m, --mst=MST         Minimum spanning tree algorithm to use:\n"
	  "                        prim-bin    Prim's, binary heap\n"
	  "                        prim-fib    Prim's, Fibonacci heap\n"
	  "-R, --rng=RNG         Random number generator to use:\n"
	  "                        mt          Mersenne Twister (default)\n"
	  "                        std         C library\n"
	  "-p, --prox=DIST       Divide points into two distributions\n"
	  "                      of same type, separated by distance DIST\n"
	  "                      (default is only one distribution)\n"
	  "-r, --runs=RUNS       Number of runs (default 16).\n"
	  "-s, --seed=SEED       Initial random seed (time-based default).\n"
	  "-h, --help            Displays this help screen.\n"
	  "-V, --version         Reports version and copyright information.\n",
	  pgm_name);

  exit (EXIT_SUCCESS);
}

/* Parses command-line arguments from null-terminated array |args|.
   Sets up |options| appropriately to correspond. */
void
parse_command_line (char **args, struct options *options)
{
  static const struct long_option option_tab[] = 
    {
      {"dims", 1, 'd'},
      {"count", 1, 'c'},
      {"average", 1, 'a'},
      {"sig", 1, 'S'},
      {"dist", 1, 'D'},
      {"mst", 1, 'm'},
      {"rng", 1, 'R'},
      {"prox", 1, 'p'},
      {"runs", 1, 'r'},
      {"seed", 1, 's'},
      {"help", 0, 'h'},
      {"version", 0, 'V'},
      {NULL, 0, 0},
    };

  struct option_state state;

  /* Default options. */
  options->dims = 5;
  options->count = 100;
  options->average = 1;
  options->sig = .05;
  options->mst = MST_PRIM_BIN_HEAP;
  options->rng = RNG_MT;
  options->prox = 0.0;
  options->dist = DIST_UNIFORM;
  options->runs = 16;
  options->seed = time_seed () % 32768u;

  if (*args == NULL)
    return;

  option_init (&state, args + 1, "d:c:a:S:D:m:R:p:r:s:hV", option_tab);
  for (;;)
    {
      char *arg;
      int id = option_get (&state, &arg);
      if (id == -1)
	break;

      switch (id)
	{
	case 'd':
	  options->dims = stoi (arg);
	  if (options->dims < 1)
	    fail ("bad number of dimensions (%d)\n", options->dims);
	  break;

	case 'c':
	  options->count = stoi (arg);
	  if (options->count < 1)
	    fail ("bad point count (%d)\n", options->count);
	  break;

	case 'a':
	  options->average = stoi (arg);
	  if (options->average < 1)
	    fail ("bad averaging repetitions %d\n", options->average);
	  break;

	case 'S':
	  options->sig = strtod (arg, NULL);
	  if (options->sig <= 0.0 || options->sig >= 1.0)
	    fail ("significance level %g not between 0 and 1", options->sig);
	  break;

	case 'D':
	  if (!strcmp (arg, "uniform"))
	    options->dist = DIST_UNIFORM;
	  else if (!strcmp (arg, "decreasing"))
	    options->dist = DIST_DECREASING;
	  else if (!strcmp (arg, "normal"))
	    options->dist = DIST_NORMAL;
	  else
	    fail ("unknown distribution `%s'", arg);
	  break;

	case 'm':
	  if (!strcmp (arg, "prim-bin"))
	    options->mst = MST_PRIM_BIN_HEAP;
	  else if (!strcmp (arg, "prim-fib"))
	    options->mst = MST_PRIM_FIB_HEAP;
	  else
	    fail ("unknown MST algorithm `%s'", arg);
	  break;

	case 'R':
	  if (!strcmp (arg, "std"))
	    options->rng = RNG_STD;
	  else if (!strcmp (arg, "mt"))
	    options->rng = RNG_MT;
	  else
	    fail ("unknown PRNG `%s'", arg);
	  break;

	case 'p':
	  options->prox = stoi (arg);
	  if (options->prox < 0)
	    fail ("bad proximity (%g)\n", options->prox);
	  break;

	case 'r':
	  options->runs = stoi (arg);
	  if (options->runs < 1)
	    fail ("bad run count (%d)\n", options->runs);
	  break;

	case 's':
	  options->seed = stoi (arg);
	  break;

	case 'h':
	  usage ();

	case 'V':
	  fputs ("Uniformity test unit test 1.0\n"
		 "Copyright (C) 2001 Ben Pfaff.\n"
		 "Copyright (C) 2001 Free Software Foundation, Inc.\n"
		 "See COPYING.BSD and COPYING.GPL for license.\n"
		 "This program comes with NO WARRANTY, not even for\n"
		 "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
		 stdout);
	  exit (EXIT_SUCCESS);

	default:
	  assert (0);
	}
    }
}

/* Returns a normal deviate with mean 0 and standard deviation 1.
   Derived from the C FAQ. */
double
normal_deviate (struct unf_rng *rng)
{
  static double v1, v2, s;
  static int phase = 0;
  double x;

  if (phase == 0)
    {
      do
	{
	  double u1 = rng->unf_get ();
	  double u2 = rng->unf_get ();
	  
	  v1 = 2.0 * u1 - 1.0;
	  v2 = 2.0 * u2 - 1.0;
	  s = v1 * v1 + v2 * v2;
	}
      while (s >= 1.0 || s == 0.0);

      x = v1 * sqrt (-2.0 * log (s) / s);
    }
  else
    x = v2 * sqrt (-2.0 * log (s) / s);
  
  phase ^= 1;
  
  return x;
}

#ifdef __GNUC__
#define unused __attribute__((__unused__))
#else
#define unused
#endif

int
main (int argc unused, char *argv[])
{
  struct options cmdline;
  struct unf_options test_options;
  double *data;
  int nonuniform;
  int i;

  parse_command_line (argv, &cmdline);

  unf_init_options (&test_options);
  switch (cmdline.mst)
    {
    case MST_PRIM_BIN_HEAP:
      test_options.unf_mst = &unf_mst_prim_binary;
      break;

    case MST_PRIM_FIB_HEAP:
      test_options.unf_mst = &unf_mst_prim_fibonacci;
      break;

    default:
      assert (0);
    }
  switch (cmdline.rng)
    {
    case RNG_MT:
      test_options.unf_rng = &unf_rng_mt;
      break;

    case RNG_STD:
      test_options.unf_rng = &unf_rng_system;
      break;

    default:
      assert (0);
    }

  data = malloc (sizeof *data * cmdline.count * cmdline.dims);
  if (data == NULL)
    {
      printf ("memory allocation error\n");
      return EXIT_FAILURE;
    }

  printf ("Test program for uniformity modules; use --help for help.\n");

  nonuniform = 0;
  for (i = 0; i < cmdline.runs; i++)
    {
      int j, k;
      double *p;
      double sig;

      printf ("Seed = %u...  ", cmdline.seed);
      fflush (stdout);
      test_options.unf_rng->unf_seed (cmdline.seed);

      p = data;
      for (j = 0; j < cmdline.count; j++)
	switch (cmdline.dist)
	  {
	  case DIST_UNIFORM:
	    for (k = 0; k < cmdline.dims; k++)
	      *p++ = test_options.unf_rng->unf_get ();
	    break;

	  case DIST_DECREASING:
	    {
	      double prev = 1.0;
	      for (k = 0; k < cmdline.dims; k++)
		prev = *p++ = test_options.unf_rng->unf_get () * prev;
	    }
	    break;

	  case DIST_NORMAL:
	    for (k = 0; k < cmdline.dims; k++)
	      *p++ = normal_deviate (test_options.unf_rng);
	    break;
	    
	  default:
	    abort ();
	  }

      if (cmdline.prox > 0)
	{
	  for (j = 0; j < cmdline.count / 2; j++)
	    data[j * cmdline.dims] -= cmdline.prox / 2.0;
	  for (; j < cmdline.count; j++)
	    data[j * cmdline.dims] += cmdline.prox / 2.0;
	}

      sig = 0.0;
      for (j = 0; j < cmdline.average; j++)
	{
	  double one_sig = unf_test (&test_options, data,
				     cmdline.count, cmdline.dims);
	  if (one_sig == -1)
	    printf ("error!\n");
	  else
	    sig += one_sig;
	}

      printf ("s=%.4f: %s\n",
	      sig, sig > cmdline.sig ? "uniform" : "non-uniform");
      if (sig <= cmdline.sig)
	nonuniform++;

      cmdline.seed++;
    }
  printf ("%d of %d (%g%%) detected as non-uniform.\n",
	  nonuniform, cmdline.runs, nonuniform * 100.0 / cmdline.runs);
  
  return 0;
}
